//---------------------------------------------
// エミュレーション結果の表現法/インターフェース

#ifndef RENDERER_H
#define RENDERER_H

#include "gbcore/gb.h"

#define BUF_LEN 160

enum{
	SCR_X1,
	SCR_X2,
	SCR_END,
};
extern const char *scr_names[];
extern int pad_state;
extern int bMenu;

void renderer_init();
void renderer_reset();
void renderer_refresh();
void render_screen(void *buf);
byte renderer_get_time(int type);
word renderer_get_sensor(BOOL x_y);
void renderer_set_time(int type,byte dat);
void renderer_set_bibrate(BOOL bibrate);
int renderer_get_timer_state();
void renderer_set_timer_state(int timer);
void renderer_set_msg(const char msg[]);

#endif
